"""
Specialized pages for interactive experiments.
"""

import operator
import time
from abc import abstractmethod

import alfred3 as al
from alfred3 import admin
from alfred3._helper import inherit_kwargs
from alfred3.element.misc import RepeatedCallback

from ._util import NoMatch
from .element import ToggleMatchMakerActivation, ViewMembers


@inherit_kwargs
class MatchMakerActivation(admin.OperatorPage):
    """
    Admin page for toggling MatchMaker activation.

    A deactivated MatchMaker will abort the experiment when MatchMaking
    is tried. You can use :attr:`.MatchMaker.active` and
    :meth:`.MatchMaker.check_activation` to manually check for activation.
    The former simply returns a boolean value, indicating the MatchMaker's
    status. The latter one will automatically abort the experiment, if
    the MatchMaker is inactive.

    .. note:: This page requires operator access (level 2) to the admin mode.

    Args:
        matchmaker_location (str): A string that provides a way for the
            page to access a MatchMaker instance.
        {kwargs}


    Examples:

        To activate the admin mode, we need to set passwords for all three
        admin levels in *secrets.conf*::

            # secrets.conf
            [general]
            adminpass_lvl1 = demo
            adminpass_lvl2 = use-better-passwords
            adminpass_lvl3 = to-protect-access

        We add the activation page to the experiment's admin mode. Because
        we attach the MatchMaker to ``exp.plugins.mm``, we provide the
        ``"plugins.mm"`` as the value for *matchmaker_location*::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.SequentialSpec("role1", "role2", nslots=10, name="demo")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

            exp.admin += ali.MatchMakerActivation("plugins.mm", name="activate")

            exp += al.ForwardOnlySection(name="main")

            @exp.member(of_section="main")
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match()
                    self.exp.plugins.group = group
                    return True

    """

    title = "MatchMaker Activation"

    def __init__(self, matchmaker_location: str, **kwargs):
        super().__init__(**kwargs)
        self.matchmaker_location = matchmaker_location
        self.match_maker = None

    def added_to_experiment(self, exp):
        self.match_maker = operator.attrgetter(self.matchmaker_location)(exp)
        super().added_to_experiment(exp)

    def on_exp_access(self):
        self += al.Text(
            f"Matchmaker ID: {self.match_maker.matchmaker_id}", align="center"
        )
        self += ToggleMatchMakerActivation(match_maker=self.match_maker, align="center")


@inherit_kwargs
class MatchMakerMonitoring(admin.SpectatorPage):
    """
    Admin page for monitoring a MatchMaker's work.

    Displays tabular information about a MatchMaker's work.

    .. note:: This page requires spectator access (level 1) to the admin mode.

    Args:
        matchmaker_location (str): A string that provides a way for the
            page to access a MatchMaker instance.
        {kwargs}

    Examples:

        To activate the admin mode, we need to set passwords for all three
        admin levels in *secrets.conf*::

            # secrets.conf
            [general]
            adminpass_lvl1 = demo
            adminpass_lvl2 = use-better-passwords
            adminpass_lvl3 = to-protect-access

        We add the monitoring page to the experiment's admin mode. Because
        we attach the MatchMaker to ``exp.plugins.mm``, we provide the
        ``"plugins.mm"`` as the value for *matchmaker_location*::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.SequentialSpec("role1", "role2", nslots=10, name="demo")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

            exp.admin += ali.MatchMakerMonitoring("plugins.mm", name="monitor")

            exp += al.ForwardOnlySection(name="main")

            @exp.member(of_section="main")
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match()
                    self.exp.plugins.group = group
                    return True

    """

    title = "MatchMaker Monitoring"

    def __init__(self, matchmaker_location: str, **kwargs):
        super().__init__(**kwargs)
        self.matchmaker_location = matchmaker_location
        self.match_maker = None

    def added_to_experiment(self, exp):
        self.match_maker = operator.attrgetter(self.matchmaker_location)(exp)
        super().added_to_experiment(exp)

    def on_exp_access(self):
        self += al.Text(
            f"Matchmaker ID: {self.match_maker.matchmaker_id}", align="center"
        )
        self += ViewMembers(match_maker=self.match_maker, name="view_mm")
        self += al.VerticalSpace("30px")
        self += al.Text(
            "Note: The MatchMaker Admin displays only sessions for which the"
            " matchmaking process has been started.",
            font_size="small",
            width="full",
        )

        self += al.Style(code="#view_mm {{font-size: 85%;}}")

        # datatables javascript package
        # self += al.Style(url="//cdn.datatables.net/1.10.24/css/jquery.dataTables.min.css")
        self += al.Style(
            url="https://cdn.datatables.net/1.10.24/css/dataTables.bootstrap4.min.css"
        )
        self += al.JavaScript(
            url="//cdn.datatables.net/1.10.24/js/jquery.dataTables.min.js"
        )
        self += al.JavaScript(
            url="https://cdn.datatables.net/1.10.24/js/dataTables.bootstrap4.min.js"
        )


class DefaultWaitingTimeoutPage(al.Page):
    """
    Default page to be displayed upon waiting timeouts.
    """

    title = "Timeout"
    name = "default_waiting_timeout_page"

    def on_exp_access(self):
        self += al.VerticalSpace("50px")
        self += al.Html(al.icon("user-clock", size="80pt"), align="center")
        self += al.VerticalSpace("100px")
        self += al.Text("Sorry, waiting took too long.", align="center")


class DefaultWaitingExceptionPage(al.Page):
    """
    Default page to be displayed upon waiting exceptions.
    """

    title = "Experiment aborted"
    name = "default_waiting_exception_page"

    def on_exp_access(self):
        self += al.VerticalSpace("50px")
        self += al.Html(al.icon("user-clock", size="80pt"), align="center")
        self += al.VerticalSpace("100px")
        self += al.Text(
            "Sorry, the experiment was aborted while waiting.", align="center"
        )


@inherit_kwargs
class WaitingPage(al.NoNavigationPage):
    """
    A page that provides a waiting screen for synchronization in
    interactive experiments.

    This page is must be derived. You define the condition by defining
    the :meth:`.wait_for` method. The page has a default design, set in
    :meth:`.on_exp_access` - you can safely redefine this method to use
    your own design.

    Once the :meth:`.wait_for` method returns a *True*-like value (i.e.,
    a value for which ``bool(value) == True`` holds), the
    WaitingPage will proceed to the next page automatically.

    .. important:: The :meth:`.wait_for` method *must* return a value,
        otherwise you will always run into the timeout.

    Args:
        wait_msg (str): Text to be displayed in the default layout.
            Defaults to None, in which case the text is
            "Waiting for other group members." Can be defined as a
            class attribute.
        wait_timeout (int): Maximum waiting time in seconds. If *wait_for*
            does not return a *True*-like value within waiting time, the experiment
            session is aborted. Defaults to None, in which case the
            timeout is ``60 * 20``, i.e. 20 minutes.
            Can be defined as a class attribute.
        wait_sleep_time (int): Number of seconds in between two internal
            calls to :meth:`.wait_for`. Defaults to None, in which case
            a call will be made every two seconds.
            Can be defined as a class attribute.
        wait_timeout_page (alfred3.Page): A custom page to be displayed
            in case of a waiting timeout. Defaults to an instance of
            :class:`.DefaultWaitingTimeoutPage`.
            Can be defined as a class attribute.
        wait_exception_page (alfred3.Page): A custom page to be displayed
            in case of an exception during waiting. Defaults to an
            instance of :class:`.DefaultWaitingExceptionPage`.
            Can be defined as a class attribute.

        {kwargs}

    Examples:

        How to use the WaitingPage for matching::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.ParallelSpec("a", "b", nslots=10, name="spec1")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

            @exp.member
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match()
                    self.exp.plugins.group = group
                    return True

            @exp.member
            class Success(al.Page):

                def on_first_show(self):
                    group = self.exp.plugins.group
                    role = group.me.role

                    self += al.Text(f"Successfully matched to role: {{role}}")


        The example demonstrates how to use the waiting page in
        an ongoing experiment to wait until a group member has proceeded
        to a certain point in the experiment (more precisely, until a
        group member has filled a certain input element in this case)::


            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.ParallelSpec("a", "b", nslots=10, name="spec1")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)


            @exp.member
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match()
                    self.exp.plugins.group = group
                    return True


            @exp.member
            class InputPage(al.Page):

                def on_exp_access(self):
                    self += al.TextEntry(leftlab="Enter text", name="el1", force_input=True)


            @exp.member
            class Sync1(ali.WaitingPage):

                def wait_for(self):
                    you = self.exp.plugins.group.you
                    return you.values.get("el1", False)


            @exp.member
            class SyncSuccess(al.Page):

                def on_first_show(self):
                    self += al.Text("Successfully synced.")

    """

    #: str: Page title
    title = "Waiting"

    #: Text to be displayed in the default layout.
    #: Defaults to None, in which case the text is
    #: "Waiting for other group members."
    wait_msg: str = "Waiting for other group members."

    #: Maximum waiting time in seconds. If *wait_for*
    #: does not return a *True*-like value within waiting time, the experiment
    #: session is aborted. Defaults to None, in which case the
    #: timeout is ``60 * 20``, i.e. 20 minutes.
    wait_timeout: int = 60 * 20

    #: Number of seconds in between two internal
    #: calls to :meth:`.wait_for`. Defaults to None, in which case
    #: a call will be made every two seconds.
    wait_sleep_time: int = 3

    #: Abort page to be displayed on timeout
    wait_timeout_page = None

    #: Abort page to be displayed on other exceptions during waiting
    wait_exception_page = None

    def __init__(
        self,
        *args,
        wait_msg: str = None,
        wait_timeout: int = None,
        wait_sleep_time: int = None,
        wait_timeout_page: al.Page = None,
        wait_exception_page: al.Page = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if wait_msg is not None:
            self.wait_msg = wait_msg

        if wait_timeout is not None:
            self.wait_timeout = wait_timeout

        if wait_sleep_time is not None:
            self.wait_sleep_time = wait_sleep_time

        if wait_timeout_page is not None:
            self.wait_timeout_page = wait_timeout_page
        else:
            self.wait_timeout_page = DefaultWaitingTimeoutPage()

        if wait_exception_page is not None:
            self.wait_exception_page = wait_exception_page
        else:
            self.wait_exception_page = DefaultWaitingExceptionPage()

        self += RepeatedCallback(
            func=self._wait_for,
            interval=self.wait_sleep_time,
            followup="custom",
            custom_js="if (data) {move('forward')};",
        )

        #: Time of waiting start in seconds since epoch
        #: We are not using the first show time here, because we want
        #: to try to synchronize the waiting timeout with the displayed
        #: Counter as best as possible. If we used the first show time,
        #: that may be an earlier time point.
        self.waiting_start: float = None
        self.countup = al.CountUp(font_size=30, align="center")

    @property
    def expiration_time(self) -> float:
        """
        float: Point in time at which the waiting page expires in
        seconds since epoch.
        """
        if not self.waiting_start:
            return None

        return self.waiting_start + self.wait_timeout

    @property
    def expired(self) -> bool:
        """
        bool: *True* if the page has exceeded its maximum waiting time.
        """
        if self.expiration_time:
            now = time.time()
            return now > self.expiration_time
        else:
            return False

    @property
    def passed_time(self) -> float:
        return time.time() - self.waiting_start

    @abstractmethod
    def wait_for(self):
        """
        Once this method returns *True*, the page automatically
        forwards participants to the next page.

        It will be repeatedly called internally with the time between
        two calls defined by :attr:`.wait_sleep_time`.
        """
        pass

    def _wait_for(self):
        """
        This method gets called repeatedly via ajax callback from the
        waiting page client-side. If it returns *True*, the experiment will move
        on to the next page. Otherwise, the callback will try again
        until the timeout is reached.
        """

        if not self.waiting_start:
            self.waiting_start = time.time()
            self.countup.start_time = self.waiting_start

        if self.expired:
            self.on_expire()
            return True

        try:
            wait_status = self.wait_for()

        except NoMatch:
            return False  # return False so that the repeated callback will try again

        except Exception:
            self.log.exception("Exception in waiting function. Aborting experiment.")
            self.exp.abort(reason="waiting error", page=self.wait_exception_page)
            return True

        return wait_status

    def on_expire(self):
        self.log.exception("Timeout on waiting page. Aborting experiment.")
        self.exp.abort(reason="timeout", page=self.wait_timeout_page)

    def on_exp_access(self):
        spinning_icon = al.icon("spinner", spin=True, size="90pt")

        self += al.VerticalSpace("100px")
        self += al.Text(spinning_icon, align="center")
        self += al.VerticalSpace("30px")
        self += self.countup
        self += al.Text(self.wait_msg, align="center")


@inherit_kwargs
class MatchingPage(WaitingPage):
    """
    A waiting page for matchmaking.

    Args:
        {kwargs}

    Examples:
        ::

            import alfred3 as al
                import alfred3_interact as ali

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    spec = ali.SequentialSpec("a", "b", nslots=10, name="spec1")
                    exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

                @exp.member
                class Match(ali.MatchingPage):

                    def wait_for(self):
                        group = self.exp.plugins.mm.match()
                        self.exp.plugins.group = group
                        return True

                @exp.member
                class Success(al.Page):

                    def on_first_show(self):
                        group = self.exp.plugins.group
                        role = group.me.role

                        self += al.Text(f"Successfully matched to role: {{role}}")
    """


@inherit_kwargs
class MatchTestPage(al.Page):
    """
    A page for testing interactive experiments during development.

    Args:
        {kwargs}

    The page expects a group object to be locatod at
    ``self.exp.plugins.group`` when it is first shown.
    It displays some information about the group, as well as a group chat.

    Examples:
        An example experiment with two specs with three roles each,
        random matching and a test page::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec1 = ali.ParallelSpec("a", "b", "c", nslots=10, name="spec1")
                spec2 = ali.ParallelSpec("a2", "b2", "c2", nslots=10, name="spec2")
                exp.plugins.mm = ali.MatchMaker(spec1, spec2, exp=exp)


            @exp.member
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match_random()
                    self.exp.plugins.group = group
                    return True

            exp += ali.MatchTestPage(name="test")


    """

    title = "MatchMaking Test Page"

    def on_first_show(self):
        group = self.exp.plugins.group
        role = group.me.role

        self += al.Text("## This session")
        self += al.Text(f"Successfully matched to role: {role}")
        self += al.Text(f"This member's spec name: {group.spec_name}")

        self += al.VerticalSpace("15pt")
        self += al.Hline()
        self += al.VerticalSpace("15pt")

        self += al.Text("## Group Members")
        for m in group.members():
            self += al.Text(f"Member with role '{m.role}': {m.data.session_id}")

        self += al.VerticalSpace("15pt")
        self += al.Hline()
        self += al.VerticalSpace("15pt")

        self += al.Text("## Group Specs")
        for m in group.members():
            self += al.Text(
                f"Spec of member with role '{m.role}': {m.mm.group.spec_name}"
            )

        self += al.VerticalSpace("15pt")
        self += al.Hline()
        self += al.VerticalSpace("15pt")

        self += al.Text("## Group Chat")
        self += group.chat()



from . import MatchMaker, NoMatch, ParallelSpec, SequentialSpec


def get_group(
    exp, roles: list = None, ongoing_sessions_ok: bool = False, nslots: int = 5
):
    roles = ["a", "b"] if roles is None else roles
    spec = SequentialSpec(
        *roles, nslots=nslots, name="test", ongoing_sessions_ok=ongoing_sessions_ok
    )
    mm = MatchMaker(spec, exp=exp)
    group = mm.match_to("test")
    return group


def get_group_groupwise(exp_factory: callable, roles: list = None, nslots: int = 5):
    roles = roles if roles is not None else ["a", "b"]
    exp1 = exp_factory()
    exp2 = exp_factory()

    spec1 = ParallelSpec(*roles, nslots=nslots, name="test")
    spec2 = ParallelSpec(*roles, nslots=nslots, name="test")

    mm1 = MatchMaker(spec1, exp=exp1)
    mm2 = MatchMaker(spec2, exp=exp2)

    try:
        group1 = mm1.match_to("test")
    except NoMatch:
        pass

    mm2.match_to("tes")
    group1 = mm1.match_to("test")

    return group1


def get_multiple_groups(*mm, specname: str = "test"):

    try:
        mm[0].match_to(specname)
    except NoMatch:
        pass

    groups = [m.match_to(specname) for m in reversed(mm)]

    return tuple(reversed(groups))



import json
import time
import typing as t
from dataclasses import dataclass, field
from typing import Iterator, List

from alfred3._helper import inherit_kwargs
from alfred3.data_manager import DataManager as dm
from alfred3.data_manager import saving_method
from alfred3.quota import QuotaData, SessionQuota, Slot, SlotManager

from .group import GroupType


@dataclass
class GroupSlot:
    label: str
    group_ids: List[str] = field(default_factory=list)

    def get_data(self, exp) -> dict:
        if saving_method(exp) == "local":
            return self._get_data_local(exp)
        elif saving_method(exp) == "mongo":
            return self._get_data_mongo(exp)

    def _get_data_mongo(self, exp) -> t.Iterator[dict]:
        return exp.db_misc.find({"group_id": {"$in": self.group_ids}})

    def _get_data_local(self, exp) -> t.Iterator[dict]:
        path = exp.config.get("interact", "path", fallback="save/interact")
        path = exp.subpath(path)

        for fp in path.iterdir():
            if "group" not in str(fp):
                continue
            with open(fp, encoding="utf-8") as f:
                group_data = json.load(f)
            if group_data["group_id"] in self.group_ids:
                yield group_data

    def most_recent_save(self, exp) -> float:
        data = self.get_data(exp)
        oldest_saves = []
        # for each group, we count the oldest activity of pending members
        # because this is the "weakest link"
        # Thus, we get the newest save time of the slowest member of each group
        for group_data in data:
            # query for active sessions
            members = self._pending_members(exp, group_data, ["exp_save_time"])
            save_times = [m["exp_save_time"] for m in members]
            oldest_saves.append(min(save_times))

        # from the oldest activities, we take the newest one
        return max(oldest_saves)

    def _pending_members(
        self, exp, group_data, projection: t.Union[dict, list] = None
    ) -> t.Iterator[dict]:
        if saving_method(exp) == "local":
            return self._pending_members_local(exp, group_data)
        elif saving_method(exp) == "mongo":
            return self._pending_members_mongo(exp, group_data, projection)

    def _pending_members_local(self, exp, group_data) -> t.Iterator[dict]:
        earliest_start = time.time() - exp.session_timeout
        path = exp.config.get("local_saving_agent", "path")
        path = exp.subpath(path)
        for session_data in dm.iterate_local_data(dm.EXP_DATA, path):
            in_group = session_data["exp_session_id"] in group_data["members"]
            aborted = session_data["exp_aborted"]
            finished = session_data["exp_finished"]
            start = session_data["exp_start_time"]
            expired = False if start is None else start < earliest_start
            if in_group and not aborted and not finished and not expired:
                yield session_data

    def _pending_members_mongo(
        self, exp, group_data, projection: t.Union[dict, list] = None
    ) -> t.Iterator[dict]:
        earliest_start = time.time() - exp.session_timeout

        q = {
            "type": dm.EXP_DATA,
            "exp_session_id": {"$in": group_data["members"]},
            "exp_finished": False,
            "exp_aborted": False,
            "$or": [
                {"exp_start_time": {"$gte": earliest_start}},
                {"exp_start_time": None},
            ],
        }

        return exp.db_main.find(q, projection=projection)

    def npending(self, exp) -> int:
        counts = self._npending_members(exp)
        npending_groups = len([gid for gid, npending in counts.items() if npending > 0])
        return npending_groups

    def _npending_members(self, exp) -> dict:
        if saving_method(exp) == "local":
            counts = self._npending_groups_local(exp)
        elif saving_method(exp) == "mongo":
            counts = self._npending_groups_mongo(exp)

        return counts

    def _count_pending(self, exp, members, data: list, cursor: t.Iterator) -> dict:
        earliest_start = time.time() - exp.session_timeout
        gids = [d["group_id"] for d in data]

        counts = {gid: 0 for gid in gids}
        exited = {gid: False for gid in gids}

        for session_data in cursor:
            aborted = session_data["exp_aborted"]
            finished = session_data["exp_finished"]
            start = session_data["exp_start_time"]
            expired = False if start is None else start < earliest_start

            gid = members.get(session_data["exp_session_id"], None)

            if gid and not aborted and not finished and not expired:
                counts[gid] += 1

            if "groupwise" in data[0]["type"] and (aborted or expired):
                exited[gid] = True

        return counts

    def _npending_groups_local(self, exp) -> int:
        data = self.get_data(exp)
        data = list(data)

        path = exp.config.get("local_saving_agent", "path")
        path = exp.subpath(path)
        cursor = dm.iterate_local_data(dm.EXP_DATA, path)

        members = {}
        for group_data in data:
            members.update(
                {sid: group_data["group_id"] for sid in group_data["members"]}
            )

        counts = self._count_pending(exp, members, data, cursor)

        return counts

    def _npending_groups_mongo(self, exp) -> int:
        p = ["group_id", "members", "type"]
        data = exp.db_misc.find({"group_id": {"$in": self.group_ids}}, projection=p)
        data = list(data)
        members = {}
        for group_data in data:
            members.update(
                {sid: group_data["group_id"] for sid in group_data["members"]}
            )

        q = {"exp_session_id": {"$in": list(members)}, "type": dm.EXP_DATA}
        p = ["exp_finished", "exp_aborted", "exp_start_time", "exp_session_id"]
        cursor = exp.db_main.find(q, p)
        counts = self._count_pending(exp, members, data, cursor)

        return counts

    def _finished_members(
        self, exp, group_data: dict, projection: t.Union[dict, list]
    ) -> t.Iterator[dict]:
        if saving_method(exp) == "local":
            return self._finished_members_local(exp, group_data)
        elif saving_method(exp) == "mongo":
            return self._finished_members_mongo(exp, group_data, projection)

    def _finished_members_local(self, exp, group_data: dict) -> t.Iterator[dict]:
        path = exp.config.get("local_saving_agent", "path")
        path = exp.subpath(path)
        for session_data in dm.iterate_local_data(dm.EXP_DATA, path):
            sid = session_data["exp_session_id"]
            finished = session_data["exp_finished"]
            if sid in group_data["members"] and finished:
                yield session_data

    def _finished_members_mongo(
        self, exp, group_data: dict, projection: t.Union[dict, list]
    ) -> t.Iterator[dict]:
        q = {
            "type": dm.EXP_DATA,
            "exp_session_id": {"$in": group_data["members"]},
            "exp_finished": True,
        }

        return exp.db_main.find(q, projection=projection)

    def _nfinished_members(self, exp):
        data = self.get_data(exp)

        counts = {}
        for group_data in data:
            cursor = self._finished_members(exp, group_data, ["exp_finished"])
            nfinished = sum(session["exp_finished"] for session in cursor)
            gid = group_data["group_id"]
            counts[gid] = nfinished

        return counts

    def finished(self, exp) -> bool:
        data = self.get_data(exp)

        for group_data in data:
            cursor = self._finished_members(exp, group_data, ["exp_finished"])
            nfinished = sum(session["exp_finished"] for session in cursor)
            nroles = len(group_data["roles"])

            if nfinished == nroles:
                return True

        return False

    def pending(self, exp) -> bool:
        return not self.finished(exp) and not self.open

    @property
    def open(self) -> bool:
        return not self.group_ids

    def __contains__(self, group_ids: List[str]) -> bool:
        if self.group_ids is None:
            return False
        group_in_slot = [gid in self.group_ids for gid in group_ids]
        return any(group_in_slot)


@dataclass
class GroupSlotManager:
    slots: List[dict]

    def __post_init__(self):
        self.slots = [GroupSlot(**slot_data) for slot_data in self.slots]

    def open_slots(self, exp) -> Iterator[Slot]:
        return (slot for slot in self.slots if slot.open)

    def pending_slots(self, exp) -> Iterator[Slot]:
        return (slot for slot in self.slots if slot.pending(exp))

    # def incomplete_slots(self, exp) -> Iterator[Slot]:
    #     return (slot for slot in self.slots if slot.contains_incomplete_group(exp))

    def find_slot(self, group_ids: List[str]) -> Slot:
        for slot in self.slots:
            if group_ids in slot:
                return slot

    def next_pending(self, exp) -> Slot:
        """
        This method returns the next pending slot that is
        most likely not to finish the experiment.

        1. If there is only one pending slot, this slot is returned
        2. The slot with the least number of pending groups operating
           in it is returned.
        3. If there are multiple slots with equal numbers of pending
           groups, the method returns the slot where the most recent
           activity of the slowest member is oldest - because this slot
           is most likely to need extra groups.
        """
        slots = list(self.pending_slots(exp))

        if len(slots) == 1:
            return slots[0]

        slots = self._sparsest_slots(slots, exp)

        if len(slots) == 1:
            return slots[0]

        return self._oldest_slot(slots, exp)

    def _sparsest_slots(self, slots, exp) -> List[Slot]:

        npending = [slot.npending(exp) for slot in slots]
        n = min(npending)
        minimal_pending = [slot for slot in slots if slot.npending(exp) == n]
        return minimal_pending

    def _oldest_slot(self, slots, exp) -> Slot:
        most_recent_save = [slot.most_recent_save(exp) for slot in slots]
        oldest = min(most_recent_save)
        i = most_recent_save.index(oldest)
        return slots[i]


@inherit_kwargs(exclude=["session_ids"])
class ParallelGroupQuota(SessionQuota):
    """
    Manages quota for parallel groups.

    Args:
        {kwargs}
    """

    group_type = GroupType.PARALLEL

    def __init__(self, nslots: int, exp, name: str = "group_quota", **kwargs):
        if kwargs.pop("session_ids", False):
            raise ValueError("Unsupported argument: 'session_ids'")

        super().__init__(nslots=nslots, exp=exp, name=name, **kwargs)
        self.group_id = None

    @staticmethod
    def _use_comptability(**kwargs) -> bool:
        return False

    def count(self, group, raise_exception: bool = False):
        if not group.data.group_type == self.group_type:
            raise ValueError(f"Group type {group.gtype} != {self.group_type}")
        self.group_id = group.data.group_id
        self.session_ids = group.data.members

        return super().count(raise_exception)

    def _slot_manager(self, data: QuotaData):
        try:
            return SlotManager(data.slots)
        except TypeError as e:
            group_ids = [slot.get("group_ids", False) for slot in data.slots]
            if any(group_ids):
                raise ValueError(
                    "Tried to initialize a quota for a parallel spec with data for a"
                    " sequential spec quota. This can occur, if you use the same name"
                    " for different specs. Please make sure that all names are unique."
                )
            else:
                raise e


@inherit_kwargs(exclude=["session_ids"])
class SequentialGroupQuota(ParallelGroupQuota):
    """
    Manages quota for sequential groups.

    Args:
        {kwargs}
    """

    group_type = GroupType.SEQUENTIAL

    def _update_slot(self, slot):
        return slot.group_ids.append(self.group_id)

    def _own_slot(self, data: QuotaData):
        slot_manager = self._slot_manager(data)
        slot = slot_manager.find_slot([self.group_id])
        return slot

    def _slot_manager(self, data: QuotaData):
        try:
            return GroupSlotManager(data.slots)
        except TypeError as e:
            session_groups = [slot.get("session_groups", False) for slot in data.slots]
            if any(session_groups):
                raise ValueError(
                    "Tried to initialize a quota for a sequential spec with data for a"
                    " parallel spec quota. This can occur, if you use the samename for"
                    " different specs. Please make sure that all names are unique."
                )
            else:
                raise e

    @property
    def full(self) -> bool:
        """
        bool: *True*, if all slots are taken *and* no group currently
        takes members.
        """
        with self.io as data:
            nopen = self._nopen(data)

            if nopen > 0:
                return False

            npending = self._npending(data)

            # no pending and no open slots means that the quota is allfinished
            if npending == 0:
                return True

            # pending is > 0, inclusive quota will allow new groups to be formed
            if self.inclusive:
                return False

            # Exclusive quota with pending slots and all groups in
            # those slots are complete. Thus, the quota is full
            return True


class MetaQuota:
    """
    Gives access to aggregated information about open, pending, and
    finished slots of multiple individual quotas.

    Args:
        *quotas: Variable number of quota
    """

    def __init__(self, *quotas):
        self.quotas = quotas

    @property
    def nopen(self) -> int:
        """
        int: Number of open slots.

        A slot is open, if there is no
        active or finished experiment session (or group of experiment
        sessions) associated with this slot.
        """
        return sum(quota.nopen for quota in self.quotas)

    @property
    def npending(self) -> int:
        """
        int: Number of pending slots.

        A slot is pending, if there is currently an active session
        (or group of sessions) associated with this slot.
        """
        return sum(quota.npending for quota in self.quotas)

    @property
    def nslots(self) -> int:
        """
        int: Total number of available slots.
        """
        return sum(quota.nslots for quota in self.quotas)

    @property
    def nfinished(self) -> int:
        """
        int: Number of finished slots.

        A slot is finished, if at least one session (or group of
        sessions in case of group quotas) associated with this slot has
        finished the experiment.
        """
        return sum(quota.nfinished for quota in self.quotas)

    @property
    def allfinished(self) -> bool:
        """
        bool: *True*, if all slots are finished.

        A slot is finished, if at least one session (or group of
        sessions in case of group quotas) associated with this slot has
        finished the experiment.
        """
        return all([quota.allfinished for quota in self.quotas])

    @property
    def full(self) -> bool:
        """
        bool: *True*, if all quotas are full.

        A quota is full, if there are no slots available to assign to
        sessions or groups.
        """
        return all([quota.full for quota in self.quotas])



"""
Functionality related to groups.
"""
import json
import random
import time
from collections import UserDict
from dataclasses import asdict, dataclass, field
from itertools import chain
from pathlib import Path
from traceback import format_exception
from typing import Iterator, List
from uuid import uuid4

from pymongo.collection import ReturnDocument

from ._util import BusyGroup, MatchingError, saving_method
from .element import Chat
from .member import GroupMember, MemberManager


class GroupType:
    SEQUENTIAL = "sequential_group"
    PARALLEL = "parallel_group"


@dataclass
class GroupData:
    exp_id: str
    exp_version: str
    matchmaker_id: str
    roles: dict
    group_type: str
    spec_name: str
    group_id: str = field(default_factory=lambda: "group-" + uuid4().hex)
    members: list = field(default_factory=list)
    timestamp: float = field(default_factory=time.time)
    active: bool = True
    busy: bool = False
    shared_data: dict = field(default_factory=dict)
    type: str = "match_group"


class SharedGroupData(UserDict):
    """
    Shared group data dictionary.

    This dictionary can be used just like a normal dictionary.
    Its benefit is that it synchronises data to a database.
    This way, the data is always automatically shared between all group
    members.
    """

    _ACCEPTED = (int, float, str, tuple, bool)

    def __init__(self, *args, group, **kwargs):
        super().__init__(*args, **kwargs)
        self.group = group
        self._db = self.group.exp.db_misc

        if self.group.data.shared_data:
            self.data = self.group.data.shared_data

        self.data["__group_id"] = self.group.data.group_id

    def _push_remote(self):
        doc = self._db.find_one_and_update(
            filter={"group_id": self.group.data.group_id},
            update={"$set": {"shared_data": self.data}},
            projection={"shared_data": True, "_id": False},
            return_document=ReturnDocument.AFTER,
        )
        if doc is not None:
            self.data = doc["shared_data"]

    def _fetch_remote(self):
        doc = self._db.find_one(
            filter={"group_id": self.group.data.group_id},
            projection={"shared_data": True, "_id": False},
        )
        if doc is not None:
            self.data = doc["shared_data"]

    def _push_local(self):
        self.group.data.shared_data = self.data
        self.group.io.save()

    def _fetch(self):
        if saving_method(self.group.exp) == "mongo":
            self._fetch_remote()

        self.data["__last_access"] = time.time()
        self._push()

    def _push(self):
        if saving_method(self.group.exp) == "mongo":
            self._push_remote()
        elif saving_method(self.group.exp) == "local":
            self._push_local()

    def last_change(self, format: str = "%Y-%m-%d, %X") -> str:
        return time.strftime(format, time.localtime(self.data["__last_change"]))

    def last_access(self, format: str = "%Y-%m-%d, %X") -> str:
        return time.strftime(format, time.localtime(self.data["__last_access"]))

    def _validate(self, key, value):

        if isinstance(value, self._ACCEPTED):
            return True

        else:
            msg = (
                f"You are setting or getting item '{key}', which refers to an object of"
                f" type {type(value)}. This is most likely a mutable type. Using a"
                " mutable object in SharedGroupData is very dangerous. You may end up"
                " losing data, because SharedGroupData can only write to the database"
                " when an item is set, not when it is changed. Please switch to using"
                " only immutable types, like tuple, str, int, float, and bool."
            )
            self.group.exp.log.warning(msg)

    def __getitem__(self, key):
        self._fetch()
        item = super().__getitem__(key)
        self._validate(key, item)
        return item

    def __setitem__(self, key, item):
        self._fetch()
        super().__setitem__(key, item)
        self.data["__last_change"] = time.time()
        self._push()
        self._validate(key, item)

    def __delitem__(self, key):
        self._fetch()
        super().__delitem__(key)
        self.data["__last_change"] = time.time()
        self._push()


class GroupHelper:
    def __init__(self, group):
        self.group = group
        self.mm = group.mm
        self.exp = group.exp


class GroupIO(GroupHelper):
    def __init__(self, group):
        super().__init__(group)
        self.saving_method = saving_method(self.exp)

    @property
    def db(self):
        return self.exp.db_misc

    @property
    def data(self):
        return self.group.data

    @property
    def query(self) -> dict:
        q = {}
        q["type"] = self.group.data.type
        q["group_id"] = self.group.data.group_id
        return q

    @property
    def path(self) -> Path:
        """
        Path: Path to group data file in offline group experiments.
        """
        parent = self.mm.io.path.parent
        return parent / f"group_{self.data.group_id}.json"

    def insert(self):
        if self.saving_method == "mongo":
            self._insert_mongo(asdict(self.data))
        elif self.saving_method == "local":
            self._insert_local(asdict(self.data))

    def _insert_mongo(self, insert: dict):
        self.db.find_one_and_update(self.query, {"$setOnInsert": insert}, upsert=True)

    def _insert_local(self, insert: dict):
        if not self.path.is_file():
            self._save_local(insert)

    def save(self):
        if self.saving_method == "mongo":
            self._save_mongo(asdict(self.data))
        elif self.saving_method == "local":
            self._save_local(asdict(self.data))

    def _save_mongo(self, data: dict):
        self.db.find_one_and_update(self.query, {"$set": data}, upsert=True)

    def _save_local(self, data: dict):
        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=4)

    def load(self) -> GroupData:

        if self.saving_method == "mongo":
            data = self._load_mongo()
        elif self.saving_method == "local":
            data = self._load_local()

        if data:
            return GroupData(**data)

    def _load_mongo(self) -> dict:
        return self.db.find_one(self.query, {"_id": False})

    def _load_local(self) -> dict:
        with open(self.path, encoding="utf-8") as fp:
            return json.load(fp)

    def load_markbusy(self) -> GroupData:
        if self.saving_method == "mongo":
            data = self._load_markbusy_mongo()
        elif self.saving_method == "local":
            data = self._load_markbusy_local()

        if data:
            return GroupData(**data)

    def _load_markbusy_mongo(self) -> dict:
        q = self.query
        q["busy"] = False
        data = self.db.find_one_and_update(
            q,
            {"$set": {"busy": True}},
            return_document=ReturnDocument.AFTER,
            projection={"_id": False},
        )

        return data

    def _load_markbusy_local(self):
        data = self._load_local()
        if not data["busy"]:
            data["busy"] = True
            self._save_local(data)
            return data
        else:
            return None

    def release(self):
        if self.saving_method == "mongo":
            self._release_mongo()
        elif self.saving_method == "local":
            self._release_local()

        self.group.data.busy = False

    def _release_mongo(self):
        self.db.find_one_and_update(self.query, {"$set": {"busy": False}})

    def _release_local(self):
        data = self._load_local()
        data["busy"] = False
        self._save_local(data)


class GroupRoles(GroupHelper):
    @property
    def manager(self):
        return self.group.manager

    @property
    def roles(self):
        return self.data.roles

    @property
    def data(self):
        return self.group.data

    def roles_of(self, sessions: str) -> Iterator[str]:
        roles = (role for role, sid in self.roles.items() if sid in sessions)
        return roles

    def finished(self) -> Iterator[str]:
        sessions = self.manager.find_finished_sessions(self.data.members)
        return self.roles_of(sessions)

    def pending(self) -> Iterator[str]:
        sessions = self.manager.find_active_sessions(self.data.members)
        return self.roles_of(sessions)

    def open(self) -> Iterator[str]:
        finished = self.manager.find_finished_sessions(self.data.members)
        pending = self.manager.find_active_sessions(self.data.members)
        sessions = list(chain(finished, pending))
        roles = (role for role, sid in self.roles.items() if sid not in sessions)
        return roles

    def assign(self, role: str, member: GroupMember):
        self.group.data.roles[role] = member.data.session_id
        member.data.role = role
        self.group.io.save()

    def next(self) -> str:
        role = next(self.open(), None)
        return role

    def shuffle(self):
        role_list = list(self.data.roles.items())
        random.shuffle(role_list)
        self.data.roles = dict(role_list)
        self.group.io.save()


class GroupMemberManager(GroupHelper):
    def __init__(self, group):
        super().__init__(group)
        self._me = None

    @property
    def manager(self):
        return self.group.manager

    @property
    def data(self):
        return self.group.data

    @property
    def me(self) -> GroupMember:
        if not self._me:
            me = self.manager.find([self.exp.session_id])
            self._me = next(me)

        return self._me

    @property
    def you(self) -> GroupMember:
        """
        GroupMember: :class:`.GroupMember`  object for
        the *other* participant in a dyad (i.e. a two-member-group).
        """
        if len(self.data.roles) > 2:
            raise MatchingError(
                "Can't use 'you' for groups with more than 2 slots, because it is"
                " ambiguous. Use the 'other_members()' generator instead, or access"
                " members based on their roles."
            )

        you = next(self.finished_other_members(), None)
        if you:
            return you

        you = next(self.active_other_members(), None)
        if you:
            return you

        you = next(self.other_members(), None)
        if you:
            return you

    @property
    def nactive(self) -> int:
        self.group.data = self.group.io.load()
        members = self.group.data.members
        active = list(self.manager.find_active_sessions(members))
        return len(active)

    @property
    def nfinished(self) -> int:
        self.group.data = self.group.io.load()
        members = self.group.data.members
        finished = list(self.manager.find_finished_sessions(members))
        return len(finished)

    def get_member_by_role(self, role: str) -> GroupMember:
        """
        GroupMember: Returns the :class:`.GroupMember` that inhabits the
        given role.
        """
        if role not in self.data.roles:
            raise AttributeError(f"Role '{role}' not found in {self}.")

        return next(self.manager.find([self.group.data.roles[role]]), None)

    def members(self) -> Iterator[GroupMember]:
        return self.manager.find(self.data.members)

    def active_members(self) -> Iterator[GroupMember]:
        sessions = list(self.manager.find_active_sessions(self.data.members))
        return self.manager.find(sessions)

    def other_members(self) -> Iterator[GroupMember]:
        for member in self.members():
            if member.data.session_id != self.exp.session_id:
                yield member

    def active_other_members(self) -> Iterator[GroupMember]:
        sessions = list(self.manager.find_active_sessions(self.data.members))
        if not sessions:
            return
        for member in self.manager.find(sessions):
            if not member.data.session_id == self.exp.session_id:
                yield member

    def finished_other_members(self) -> Iterator[GroupMember]:
        sessions = list(self.manager.find_finished_sessions(self.data.members))
        if not sessions:
            return
        for member in self.manager.find(sessions):
            if not member.data.session_id == self.exp.session_id:
                yield member

    @property
    def oldest_save(self) -> float:
        save_times = [m.info.last_save for m in self.active_members()]
        return min(save_times)


class Group:
    """
    The group object holds members and keeps track of their roles.

    The group's main task is providing access to the group members
    via their roles. Most importantly, the group offers the attributes
    :attr:`.me` and :attr:`.you`. Beyond that, you can use dot notation,
    using any member's role like an attribute on the group. The group will
    return a :class:`.GroupMember` object for the member inhabiting that
    role.

    Typically, you will not initialize a group object yourself, but
    receive it as a result of the matchmaking process via :class:`.MatchMaker`

    See Also:
        See :class:`.GroupMember` for information about member objects.

    Examples:

        In this basic example, we access both group members through their
        roles and print their inputs on the last page::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.SequentialSpec("a", "b", nslots=10, name="spec1")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

            @exp.member
            class Demo:

                def on_exp_access(self):
                    self += al.TextEntry(leftlab="Enter some text", force_input=True, name="el1")

            @exp.member
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match()
                    self.exp.plugins.group = group
                    return True

            @exp.member
            class Success(al.Page):

                def on_first_show(self):
                    group = self.exp.plugins.group
                    role = group.me.role

                    self += al.Text(f"Successfully matched to role: {role}")

                    a = group.a
                    b = group.b

                    self += al.Text(f"Values of group member a: {a.values}")
                    self += al.Text(f"Values of group member b: {b.values}")

    """

    def __init__(self, matchmaker, **data):
        self.mm = matchmaker
        self.exp = self.mm.exp

        self.data = GroupData(**self._prepare_data(data))
        self.io = GroupIO(self)
        self.manager = MemberManager(self.mm)
        self.roles = GroupRoles(self)
        self.groupmember_manager = GroupMemberManager(self)

        self._shared_data = SharedGroupData(group=self)
        self.data.shared_data = self.shared_data.data
        self.io.insert()

        self.exp.append_plugin_data_query(self._plugin_data_query)

    def _prepare_data(self, data: dict) -> dict:
        roles = data.get("roles", None)

        if not isinstance(roles, dict):
            roles = {role: None for role in roles}

        exp_id = data.get("exp_id", None)
        mm_id = data.get("matchmaker_id", None)

        data["roles"] = roles
        data["exp_id"] = exp_id if exp_id is not None else self.exp.exp_id
        data["matchmaker_id"] = mm_id if mm_id is not None else self.mm.matchmaker_id

        if "exp_version" not in data:
            data["exp_version"] = self.mm.exp_version

        data.pop("_id", None)

        return data

    @property
    def _plugin_data_query(self):
        f = {"exp_id": self.exp.exp_id, "type": self.data.type}

        q = {}
        q["title"] = "Group Data"
        q["type"] = self.data.type
        q["query"] = {"filter": f}
        q["encrypted"] = False

        return q

    @property
    def group_id(self) -> str:
        """
        str: Unique group identification number.
        """
        return self.data.group_id

    @property
    def full(self) -> bool:
        """
        bool: Indicates whether all roles in the group are filled. Counts only
        active and finished members.
        """
        nactive = self.groupmember_manager.nactive
        nfinished = self.groupmember_manager.nfinished
        return nactive + nfinished == len(self.data.roles)

    def takes_members(self, ongoing_sessions_ok: bool = False):
        """
        Indicates whether the group accepts members.

        Returns:
            bool
        """
        open_roles = len(list(self.roles.open()))
        pending_roles = len(list(self.roles.pending()))

        pending_ok = pending_roles == 0 if not ongoing_sessions_ok else True

        return bool(open_roles) and pending_ok

    @property
    def finished(self) -> bool:
        """
        bool: Indicates whether all group members have finished their
        experiment sessions.
        """
        nfinished = self.groupmember_manager.nfinished
        return nfinished == len(self.data.roles)

    @property
    def nfinished(self) -> int:
        """
        int: Number of group members who have finished their experiment
        session.
        """
        return self.groupmember_manager.nfinished

    @property
    def nactive(self) -> int:
        """
        int: Number of group members currently working on the experiment.
        """
        return self.groupmember_manager.nactive

    @property
    def shared_data(self):
        """
        DEPRECATED shared group data dictionary.
        """
        self._shared_data._fetch()
        return self._shared_data

    @property
    def me(self) -> GroupMember:
        """
        GroupMember: :class:`.GroupMember` object for the own session.
        Useful to access the own role.
        """
        return self.groupmember_manager.me

    @property
    def you(self) -> GroupMember:
        """
        GroupMember: :class:`.GroupMember`  object for
        the *other* participant in a dyad (i.e. a two-member-group).
        """
        return self.groupmember_manager.you

    @property
    def spec_name(self) -> str:
        """
        str: Name of the spec that was used to create this group.
        """
        return self.data.spec_name

    def chat(self, **kwargs) -> Chat:
        """
        Shortcut for creating a group chat.

        Args:
            **kwargs: Any argument accepted by :class:`.Chat` can be used,
                including *chat_id* and *nickname*. Specifying the latter
                two will simply override the defaults.

        See Also:
            The basic functionality is provided in the :class:`.Chat`
            element, which can be used on its own.

        Returns:
            Chat: A chat element in which the default for the chat id is
            set to the group id and the default for the nickname is set
            to each member's role.
        """
        chat_id = kwargs.pop("chat_id", self.group_id)
        nickname = kwargs.pop("nickname", self.me.role)
        return Chat(chat_id=chat_id, nickname=nickname, **kwargs)

    def members(self) -> Iterator[GroupMember]:
        """
        A generator, iterating over *all* members of the group.
        Yields :class:`.GroupMember` objects.

        Yields:
            :class:`.GroupMember`

        See Also:
            - :meth:`.active_members`
            - :meth:`.other_members`

        """
        return self.groupmember_manager.members()

    def active_members(self) -> Iterator[GroupMember]:
        """
        A generator, iterating over all *active* members of the
        group. Yields :class:`.GroupMember` objects.

        Yields:
            :class:`.GroupMember`

        See Also:
            - :attr:`.GroupMember.active`
        """
        return self.groupmember_manager.active_members()

    def other_members(self) -> Iterator[GroupMember]:
        """
        A generator, iterating over the group members
        except for :attr:`.me`. Yields :class:`.GroupMember` objects.

        Yields:
            :class:`.GroupMember`
        """
        return self.groupmember_manager.other_members()

    def active_other_members(self) -> Iterator[GroupMember]:
        """
        A generator, iterating over all *active* members of the
        group except for :attr:`.me`. Yields :class:`.GroupMember` objects.

        Yields:
            :class:`.GroupMember`

        See Also:
            - :attr:`.GroupMember.active`
        """
        return self.groupmember_manager.active_other_members()

    def __repr__(self):
        roles = str(list(self.data.roles.keys()))
        nmembers = self.groupmember_manager.nactive if self.data.members else 0
        shortid = self.data.group_id[-4:]

        return (
            f"{type(self).__name__}(roles={roles}, active members={nmembers},"
            f" id='{shortid}')"
        )

    def __getattr__(self, name):
        return self.groupmember_manager.get_member_by_role(role=name)

    def __getitem__(self, key):
        return self.groupmember_manager.get_member_by_role(role=key)

    def __eq__(self, other):
        return (type(self) == type(other)) and (
            self.data.group_id == other.data.group_id
        )

    def __iadd__(self, member):
        # TODO: Add group ID to member's experiment data

        if member.data.session_id in self.data.members:
            return self

        member.data.group_id = self.data.group_id
        self.data.members.append(member.data.session_id)
        self.io.save()
        return self

    def deactivate(self):
        with self as group:
            group.data.active = False
            group.io.save()

    def __enter__(self):
        data = self.io.load_markbusy()
        i = 0

        while not data:
            i += 1
            if i > 5:
                break
            time.sleep(1)
            data = self.io.load_markbusy()

        if not data:
            raise BusyGroup

        self.data = data
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type:
            tb = "".join(format_exception(exc_type, exc_value, tb))
            self.exp.log.error(
                f"There was an error when operating on {self}: {exc_value}."
                "The group was deactivated.\n{tb}"
            )
            self.data.active = False
            self.io.save()

        self.io.release()


class GroupManager:
    def __init__(self, matchmaker, group_type: str = None, spec_name: str = None):
        self.mm = matchmaker
        self.exp = self.mm.exp
        self.db = self.exp.db_misc
        self.saving_method = saving_method(self.exp)
        self.group_type = group_type
        self.spec_name = spec_name
        self.query = self._init_query()

    def _init_query(self):
        q = {
            "matchmaker_id": self.mm.name,
            "exp_id": self.exp.exp_id,
            "exp_version": self.mm.exp_version,
        }

        if self.spec_name is not None:
            q["spec_name"] = self.spec_name

        if self.group_type is not None:
            q["group_type"] = self.group_type

        return q

    @property
    def path(self):
        return self.mm.io.path.parent

    def groups(self) -> Iterator[Group]:
        if self.saving_method == "local":
            data = self._local_groups()
        elif self.saving_method == "mongo":
            data = self._mongo_groups()

        for gdata in data:
            return Group(self.mm, **gdata)

    def _local_groups(self) -> Iterator[dict]:
        for fpath in self.path.iterdir():
            if fpath.is_dir():
                continue

            elif not fpath.name.startswith("group"):
                continue

            else:
                with open(fpath, encoding="utf-8") as f:
                    d = json.load(f)
                    version_matches = d["exp_version"] == self.mm.exp_version
                    spec_matches = d["spec_name"] == self.spec_name
                    if version_matches and spec_matches:
                        yield d

    def _mongo_groups(self) -> Iterator[dict]:
        return self.db.find(self.query)

    def next(self, ongoing_sessions_ok: bool) -> Group:
        groups = list(self.takes_members(ongoing_sessions_ok=ongoing_sessions_ok))

        nfinished = [g.groupmember_manager.nfinished for g in groups]
        i = nfinished.index(max(nfinished))

        return groups[i]

    def takes_members(self, ongoing_sessions_ok: bool) -> Iterator[Group]:
        for group in self.active():
            if group.takes_members(ongoing_sessions_ok):
                yield group

    def active(self) -> Iterator[Group]:
        if self.saving_method == "local":
            data = self._active_local()
        elif self.saving_method == "mongo":
            data = self._active_mongo()

        for gdata in data:
            yield Group(self.mm, **gdata)

    def _active_local(self):
        for data in self._local_groups():
            if data["active"]:
                yield data

    def _active_mongo(self) -> Iterator[Group]:
        q = self.query
        q["active"] = True

        return self.db.find(q)

    def find(self, groups: List[str]) -> Iterator[Group]:
        if self.saving_method == "mongo":
            data = self._find_mongo(groups)
        elif self.saving_method == "local":
            data = self._find_local(groups)

        for gdata in data:
            return Group(self.mm, **gdata)

    def _find_mongo(self, groups: List[str]) -> Iterator[dict]:
        q = self.query
        q["group_id"] = {"$in": groups}
        return self.db.find(q)

    def _find_local(self, groups: List[str]) -> Iterator[dict]:
        for data in self._local_groups():
            if data["group_id"] in groups:
                yield data

    def find_one(self, group_id: str) -> Group:
        if self.saving_method == "mongo":
            data = self._find_one_mongo(group_id)
        elif self.saving_method == "local":
            data = self._find_one_local(group_id)

        if not data:
            return

        return Group(self.mm, **data)

    def _find_one_mongo(self, group_id: str) -> dict:
        q = self.query
        q["group_id"] = group_id
        return self.db.find_one(q)

    def _find_one_local(self, group_id: str) -> dict:
        return next(self._find_local([group_id]), None)



"""
Interfaces for defining groups.
"""

import random
import re
import typing as t
from abc import ABC, abstractmethod
from dataclasses import asdict

from ._util import MatchingError, NoMatch, saving_method
from .group import BusyGroup, Group, GroupManager, GroupType
from .member import GroupMember
from .quota import ParallelGroupQuota, SequentialGroupQuota


class SequentialMatchMaker:
    group_type = GroupType.SEQUENTIAL

    def __init__(self, *roles, matchmaker, ongoing_sessions_ok: bool, spec_name: str):
        self.roles = roles
        self.mm = matchmaker
        self.ongoing_sessions_ok = ongoing_sessions_ok

        self.log = self.mm.log
        self.group_manager = GroupManager(self.mm, self.group_type, spec_name)

        self.data = {
            "roles": self.roles,
            "group_type": self.group_type,
            "spec_name": spec_name,
        }

    @property
    def any_group_takes_members(self) -> bool:
        return any(self.group_manager.takes_members(self.ongoing_sessions_ok))

    def match(self) -> Group:
        if self.ongoing_sessions_ok and saving_method(self.mm.exp) == "local":
            raise ValueError(
                "ongoing_sessions_ok=True is not supported in local experiments."
            )

        if self.mm.member.matched:
            group = self.group_manager.find_one(self.mm.member.group_id)

        elif self.any_group_takes_members:
            try:
                group = self.match_next_group()
            except BusyGroup:
                self.log.warning("Match to existing group failed. Starting new group.")
                group = self.start_group()
        else:
            group = self.start_group()

        return group

    def start_group(self) -> Group:

        with Group(self.mm, **self.data) as group:
            self.log.info(f"Starting new group: {group}.")

            group += self.mm.member

            role = group.roles.next()
            group.roles.assign(role, self.mm.member)

            self.mm.member.io.save()
            group.io.save()
            self.log.info(
                f"Session matched to role '{self.mm.member.data.role}' in {group}."
            )
            return group

    def match_next_group(self) -> Group:
        with self.group_manager.next(self.ongoing_sessions_ok) as group:
            self.log.info(
                f"Starting stepwise match of session to existing group: {group}."
            )

            group += self.mm.member
            role = group.roles.next()
            group.roles.assign(role, self.mm.member)

            self.mm.member.io.save()
            group.io.save()

            self.log.info(
                f"Session matched to role '{self.mm.member.data.role}' in {group}."
            )
            return group


class ParallelMatchMaker:
    group_type = GroupType.PARALLEL

    def __init__(
        self, *roles, matchmaker, spec_name: str, shuffle_waiting_members: bool = True
    ):
        self.roles = roles
        self.mm = matchmaker

        self.group_manager = GroupManager(self.mm, self.group_type, spec_name)
        self.log = self.mm.log

        self.data = {
            "roles": self.roles,
            "group_type": self.group_type,
            "spec_name": spec_name,
        }

        self.shuffle_waiting_members = shuffle_waiting_members

    def match(self) -> Group:
        if saving_method(self.mm.exp) == "local":
            raise MatchingError(
                "Cannot match with parallel specs in local experiments."
            )

        with self.mm.io as data:

            if data is None:
                self.log.debug("No groupwise match conducted. MatchMaker is busy.")
            else:
                existing_group = self.get_group()
                if existing_group:
                    return existing_group

                waiting_members = self.mm.waiting_members
                enough_members_waiting = len(waiting_members) >= len(self.roles)

                if enough_members_waiting:
                    group = self.start_group(data, waiting_members)
                    return group

        raise NoMatch  # if match is not successful

    def start_group(self, data, waiting_members: t.List[GroupMember]) -> Group:

        with Group(self.mm, **self.data) as group:
            self.log.info(f"Starting new group {group}.")

            if self.shuffle_waiting_members:
                me = waiting_members[0]
                others = waiting_members[1:]
                random.shuffle(others)
                waiting_members = [me] + others

            candidates = (m for m in waiting_members)

            group.roles.shuffle()

            while not len(group.data.members) == len(group.data.roles):
                member = next(candidates)
                group += member

                role = next(group.roles.open())
                group.roles.assign(role, member)

                data.members[member.data.session_id] = asdict(member.data)

            group.io.save()
            self.mm.io.save(data=data)

            self.log.info(f"Group {group} filled. Returning group")
            return group

    def get_group(self) -> Group:
        member = self.mm.member
        member.io.load()
        if member.status.matched:
            group = self.group_manager.find_one(member.data.group_id)
            return group


class Spec(ABC):
    @abstractmethod
    def _match(self):
        pass


class ParallelSpec(Spec):
    """
    Interface for defining a parallel group.

    A parallel group is a group in which participants complete the
    experiments in parallel, i.e. they are all active simultaneously
    and can interact in real-time.

    Args:
        *roles (str): A variable number of strings, indicating which roles
            will be appointed to members of the group. The number of
            roles determines the number of group members. All roles in a
            group must be unique. Roles cannot start with numbers or
            contain spaces.
        nslots (int): Maximum number of groups that should be created
            based on this spec.
        name (str): A unique identifier for the spec.
        respect_version (bool): If True, the matchmaking will only include
            sessions that run on the same experiment version. This setting
            makes sure that there's no strange behavior if you make
            changes to an ongoing experiment.
            Defaults to True.
        inclusive (bool): If *False* (default), the quota will only assign a
            slot, if there are no pending sessions for that slot. See
            :class:`.SequentialGroupQuota` / :class:`.ParallelGroupQuota`
            for more details.
        shuffle_waiting_members (bool): If *True*, the list of waiting
            members will be shuffled before starting a new group. If
            *False*, members who have been waiting longest are
            prioritized.

    See Also:
        See :class:`.SequentialSpec` for an interface for defining a sequential group
        and :class:`.IndividualSpec` for defining an individual session spec.
    """

    group_type = GroupType.PARALLEL
    pattern = re.compile(r"^\d|\s|\.")
    _QUOTA_TYPE = ParallelGroupQuota

    def __init__(
        self,
        *roles,
        nslots: int,
        name: str,
        respect_version: bool = True,
        inclusive: bool = False,
        count: bool = True,
        shuffle_waiting_members: bool = True,
    ):

        if len(roles) != len(set(roles)):
            raise ValueError("All roles in a group must be unique.")

        self.roles = self._validate_roles(roles)
        self.name = name
        self.nslots = nslots
        self.respect_version = respect_version
        self.count = count
        self.inclusive = inclusive
        self._quota = None

        self.ongoing_sessions_ok = False

        self._shuffle_waiting_members = shuffle_waiting_members

    def _init_quota(self, exp):
        if self.count:
            self._quota = self._QUOTA_TYPE(
                nslots=self.nslots,
                exp=exp,
                respect_version=self.respect_version,
                inclusive=self.inclusive,
                name=f"{self.name}_quota",
            )

    @property
    def quota(self):
        """
        Access to the spec's :class:`.SequentialGroupQuota`,
        or :class:`.ParallelGroupQuota`,
        depending on the type of spec.
        """
        return self._quota

    def _match(self, match_maker):
        mm = ParallelMatchMaker(
            *self.roles,
            matchmaker=match_maker,
            spec_name=self.name,
            shuffle_waiting_members=self._shuffle_waiting_members,
        )

        group = mm.match()
        return group

    def _validate_roles(self, roles):
        for role in roles:

            if self.pattern.search(role):
                raise ValueError(
                    f"Error in role '{role}': Roles cannot start with numbers          "
                    "           and must not contain spaces."
                )

        return roles

    def __repr__(self):
        return f"{type(self).__name__}(roles={self.roles}, name='{self.name}')"

    def full(self, match_maker):
        return self.quota.full


class SequentialSpec(ParallelSpec):
    """
    Interface for defining a sequential group.

    A sequential group is a group in which participants complete the
    experiments not in parallel, but sequentially. That means, they
    do not interact in real-time. Instead, the first participant
    completes their session, and the next participant can access data
    from the first participant.

    Args:
        *roles (str): A variable number of strings, indicating which roles
            will be appointed to members of the group. The number of
            roles determines the number of group members. All roles in a
            group must be unique. Roles cannot start with numbers or
            contain spaces.
        nslots (int): Maximum number of groups that should be created
            based on this spec.
        name (str): A unique identifier for the spec.
        respect_version (bool): If True, the matchmaking will only include
            sessions that run on the same experiment version. This setting
            makes sure that there's no strange behavior if you make
            changes to an ongoing experiment.
            Defaults to True.
        inclusive (bool): If *False* (default), the quota will only assign a
            slot, if there are no pending sessions for that slot. See
            :class:`.SequentialGroupQuota` / :class:`.ParallelGroupQuota`
            for more details.
        ongoing_sessions_ok (bool): If False (default), new members will only
            be added to a group if all previous members of that group
            have finished their experiment sessions.

    See Also:
        See :class:`.ParallelSpec` for an interface for defining a parallel group
        and :class:`.IndividualSpec` for defining an individual session spec.
    """

    group_type = GroupType.SEQUENTIAL
    _QUOTA_TYPE = SequentialGroupQuota

    def __init__(
        self,
        *roles: str,
        nslots: int,
        name: str,
        respect_version: bool = True,
        inclusive: bool = False,
        ongoing_sessions_ok: bool = False,
        count: bool = True,
    ):
        super().__init__(
            *roles,
            nslots=nslots,
            name=name,
            respect_version=respect_version,
            inclusive=inclusive,
            count=count,
        )
        self.ongoing_sessions_ok = ongoing_sessions_ok

    def _match(self, match_maker):
        mm = SequentialMatchMaker(
            *self.roles,
            matchmaker=match_maker,
            ongoing_sessions_ok=self.ongoing_sessions_ok,
            spec_name=self.name,
        )

        group = mm.match()
        return group

    def _any_group_takes_members(self, match_maker) -> bool:
        mm = SequentialMatchMaker(
            *self.roles,
            matchmaker=match_maker,
            ongoing_sessions_ok=self.ongoing_sessions_ok,
            spec_name=self.name,
        )

        return mm.any_group_takes_members

    def full(self, match_maker) -> bool:
        any_group_takes_members = self._any_group_takes_members(match_maker)
        quota_full = self.quota.full

        return quota_full and not any_group_takes_members


class IndividualSpec(SequentialSpec):
    """
    Interface for defining an 'individual group spec'.

    The IndividualSpec can be used to include an individual condition
    in a MatchMaker speclist. The usecase arises, if you want to
    randomize the assignment of participants to at least one group
    specification and at least one individual condition.

    Args:
        nslots (int): Maximum number of sessions that should be created
            based on this spec.
        name (str): A unique identifier for the spec.
        respect_version (bool): If True, the quota will only include
            sessions that run on the same experiment version. This setting
            makes sure that there's no strange behavior if you make
            changes to an ongoing experiment.
            Defaults to True.
        inclusive (bool): If *False* (default), the quota will only assign a
            slot, if there are no pending sessions for that slot. See
            :class:`.SequentialGroupQuota` / :class:`.ParallelGroupQuota`
            for more details.

    See Also:
        See :class:`.SequentialSpec` for an interface for defining a sequential group
        and :class:`.ParallelSpec` for defining a parallel group.

    """

    def __init__(
        self,
        nslots: int,
        name: str,
        respect_version: bool = True,
        inclusive: bool = False,
        count: bool = True,
    ):
        self.roles = self._validate_roles(["individual"])
        self.name = name
        self.nslots = nslots
        self.count = count
        self.respect_version = respect_version
        self.inclusive = inclusive
        self.ongoing_sessions_ok = False



import copy
import json
import random
import time
import typing as t
from dataclasses import asdict, dataclass, field
from traceback import format_exception

from alfred3 import __version__ as alfred_version
from alfred3.alfredlog import QueuedLoggingInterface
from alfred3.exceptions import AllSlotsFull
from packaging import version
from pymongo.collection import ReturnDocument

from alfred3_interact.group import GroupManager

from ._util import MatchingError, NoMatch, saving_method
from .group import Group
from .member import GroupMember, MemberManager
from .quota import MetaQuota

ALFRED_VERSION = version.parse(alfred_version)


@dataclass
class MatchMakerData:
    exp_id: str
    exp_version: str
    matchmaker_id: str
    type: str
    members: dict = field(default_factory=dict)
    busy: bool = False
    active: bool = False
    ping_timeout: int = None


class MatchMakerIO:
    def __init__(self, matchmaker):
        self.mm = matchmaker
        self.path.parent.mkdir(exist_ok=True, parents=True)

    @property
    def db(self):
        return self.mm.exp.db_misc

    @property
    def path(self):
        p = self.mm.exp.config.get("interact", "path", fallback="save/interact")
        name = f"{self.mm.matchmaker_id}{self.mm.exp_version}.json"
        return self.mm.exp.subpath(p) / name

    @property
    def query(self):
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["exp_id"] = self.mm.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        q["matchmaker_id"] = self.mm.matchmaker_id
        return q

    def save(self, data: MatchMakerData):
        if saving_method(self.mm.exp) == "mongo":
            self._save_mongo(data)
        elif saving_method(self.mm.exp) == "local":
            self._save_local(data)
        else:
            raise MatchingError("No saving method found. Try defining a saving agent.")

    def load(self) -> MatchMakerData:
        """
        Loads MatchMakerData. If there is none, creates a MatchMakerData
        document.
        """
        if saving_method(self.mm.exp) == "mongo":
            return self._load_mongo()
        elif saving_method(self.mm.exp) == "local":
            return self._load_local()

    def load_markbusy(self) -> MatchMakerData:
        """
        Loads MatchMakerData and marks it as busy, if it is not busy
        already. In the latter case, returns None.
        """
        self.mm.busy = True
        if saving_method(self.mm.exp) == "mongo":
            return self._load_markbusy_mongo()
        elif saving_method(self.mm.exp) == "local":
            return self._load_markbusy_local()

    def release(self) -> MatchMakerData:
        """
        Releases MatchMakerData from a 'busy' state.
        Happens only in groupwise matching, so there is only a mongoDB
        version of this one.
        """
        q = copy.copy(self.query)
        q["busy"] = True
        self.db.find_one_and_update(q, {"$set": {"busy": False}})

    def _save_mongo(self, data: MatchMakerData):
        self.db.find_one_and_replace(self.query, asdict(data))

    def _save_local(self, data: MatchMakerData):
        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(asdict(data), f, sort_keys=True, indent=4)

    def _load_mongo(self):

        insert = MatchMakerData(
            exp_id=self.mm.exp.exp_id,
            exp_version=self.mm.exp_version,
            matchmaker_id=self.mm.matchmaker_id,
            active=self.mm._active,
            type=self.mm._DATA_TYPE,
            ping_timeout=self.mm.ping_timeout,
        )

        data = self.db.find_one_and_update(
            self.query,
            {"$setOnInsert": asdict(insert)},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        data.pop("_id", None)
        return MatchMakerData(**data)

    def _load_local(self) -> MatchMakerData:
        """
        If there is a matchmaker json file, this returns the data from
        that file. Otherwise, it inserts the needed json file and then
        returns the corresponding data.
        """
        if self.path.exists() and self.path.is_file():
            with open(self.path, encoding="utf-8") as f:
                return MatchMakerData(**json.load(f))
        else:
            data = MatchMakerData(
                exp_id=self.mm.exp.exp_id,
                exp_version=self.mm.exp_version,
                matchmaker_id=self.mm.matchmaker_id,
                active=self.mm._active,
                type=self.mm._DATA_TYPE,
            )

            with open(self.path, "w", encoding="utf-8") as f:
                json.dump(asdict(data), f, sort_keys=True, indent=4)
            return data

    def _load_markbusy_local(self):
        data = self._load_local()
        if not data.busy:
            data.busy = True
            self._save_local(data)
            return data
        else:
            return None

    def _load_markbusy_mongo(self):
        q = self.query
        q["busy"] = False
        data = self.db.find_one_and_update(q, {"$set": {"busy": True}})
        if data is not None:
            data.pop("_id", None)
            data["busy"] = True
            return MatchMakerData(**data)
        else:
            return None

    def __enter__(self):
        return self.load_markbusy()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self.mm.member.data.active = False
            self.mm.member.io.save()
            self.mm.exp.abort(reason="matchmaker_error")

            tb = "".join(format_exception(exc_type, exc_value, traceback))
            self.mm.exp.log.error(
                "There was an error in a locked MatchMaker operation.I deactivated the"
                f" responsible member {self.mm.member} and released the lock.\n{tb}"
            )
        self.release()
        self.mm._data = self.load()


class MatchMaker:
    """
    Creates groups of multiple experiment sessions.

    Args:
        *groupspecs: Variable number of group specifications. Each spec
            defines one blueprint for a group. Currently,
            :class:`.ParallelSpec`, :class:`.SequentialSpec`, and
            :class:`.IndividualSpec` are supported.
        exp (alfred3.ExperimentSession): Associated experiment session.
        name (str): An identifier for the matchmaker. Must be unique
            within one experiment. Can (and must) be set to a custom value, if
            you wish to use multiple matchmakers in a single experiment.
            Defaults to 'matchmaker'.
        inactive_page (alfred3.Page): Page to be displayed to participants
            if the matchmaker is inactive. If *None* (default) a default
            page will be used.
        full_page (alfred3.Page): Page to be displayed to participants
            if all specs have reached their quota of groups. If *None*
            (default) a default page will be used.
        raise_exception_if_full (bool): As an alternative to using a
            *full_page*, you can let the MatchMaker raise the
            :class:`.AllSlotsFull` exception if all specs have reached
            their quota of groups. Defaults to *False*.
        respect_version (bool): If True, the MatchMaker will only match
            sessions that run on the same experiment version into the
            same group. This setting makes sure that there's no strange
            behavior if you make changes to an ongoing experiment.
            Defaults to True.
        ping_timeout (float, int): When matching parallel groups
            (i.e. groups based on :class:`.ParallelSpec`), only active
            sessions are included in the matchmaking process. Sessions
            therefore send a signal of activity to the server on a
            regular schedule. The argument *ping_timeout* determines the
            number of seconds of inactivity before a session is
            considered to be inactive. Defaults to 15 seconds.

    The workhorse methods of the MatchMaker are :meth:`.match_random`,
    :meth:`.match_chain`, and :meth:`.match_to`. Call one of these
    methods in the :meth:`.WaitingPage.wait_for` hook to start the
    matchmaking process.

    To initialize the MatchMaker, you must first create a group
    specification, which is a sort of blueprint for building a group.
    The spec defines the size of the group, the roles that can be
    assigned to participants, and the number of groups that should be
    created based on that spec. If you supply multiple specs, you
    can randomize group creation with :meth:`.match_random` or create
    an order of priority with :meth:`.match_chain`.

    See Also:

        See :class:`.SequentialSpec` and :class:`.ParallelSpec` for more
        information on specs.

    Examples:
        A minimal example with a single spec::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                spec = ali.SequentialSpec("a", "b", nslots=10, name="spec1")
                exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

            @exp.member
            class Match(ali.WaitingPage):

                def wait_for(self):
                    group = self.exp.plugins.mm.match_to("spec1")
                    self.exp.plugins.group = group
                    return True

            @exp.member
            class Success(al.Page):

                def on_first_show(self):
                    group = self.exp.plugins.group
                    role = group.me.role

                    self += al.Text(f"Successfully matched to role: {role}")


    """

    _DATA_TYPE = "match_maker_data"

    def __init__(
        self,
        *groupspecs,
        exp,
        name: str = "matchmaker",
        inactive_page=None,
        full_page=None,
        raise_exception_if_full: bool = False,
        respect_version: bool = True,
        ping_timeout: t.Union[float, int] = 15,
    ):
        # init values
        self.exp = exp
        self.groupspecs = list(self._validate_specs(groupspecs))
        self.respect_version = respect_version
        self.exp_version = exp.version if respect_version else ""
        self.matchmaker_id = name
        self.name = name
        self.inactive_page = inactive_page
        self.full_page = full_page
        self.raise_exception_if_full = raise_exception_if_full

        # auto values
        self.spec_dict = {spec.name: spec for spec in groupspecs}
        self.log = QueuedLoggingInterface(base_logger="alfred3")
        self.log.add_queue_logger(self, __name__)
        self._active = True
        self.io = MatchMakerIO(self)
        self.member_manager = MemberManager(self)
        self.group = None
        self.member = None
        self._match_start_save = None
        self.ping_timeout = ping_timeout
        self._data = self.io.load()
        quotas = [spec.quota for spec in self.groupspecs]
        self.quota = MetaQuota(*quotas)

        self.exp.append_plugin_data_query(self._plugin_data_query)

    @property
    def _plugin_data_query(self):
        f = {}
        f = {"exp_id": self.exp.exp_id, "type": "quota_data"}

        q = {}
        q["title"] = "MatchMaker Quota Data"
        q["type"] = "quota_data"
        q["query"] = {"filter": f}
        q["encrypted"] = False

        return q

    @property
    def active(self) -> bool:
        """
        Returns *True* if the MatchMaker is active.
        """
        d = self.io.load()
        self._active = d.active
        return self._active

    @property
    def _match_start(self) -> float:
        """
        float: Start time of the matchmaking process. Should only be
        called from inside a matching function.
        """
        if self._match_start_save is None:
            self._match_start_save = time.time()

        return self._match_start_save

    @property
    def waiting_members(self) -> t.List[GroupMember]:
        """
        list: List of experiment sessions currently active and waiting
        to be matched. The sessions are represented by their
        :class:`.GroupMember` objects.
        """
        members = list(self.member_manager.waiting(self.ping_timeout))
        members = [m for m in members if not m.data.session_id == self.exp.session_id]
        if not self.member.matched:
            members = [self.member] + members
        return members

    def match(self) -> Group:
        """
        Shorthand method for conducting a match if there is only a single
        spec.

        Raises:
            ValueError: If there is more than one spec.
            NoMatch: If a single matching effort was unsuccesful. This
                exception gets handled by :class:`.WaitingPage` and is
                part of a normal matching process.

        Returns:
            Group: The group object.

        Examples:

            Matching based on a single spec::

                import alfred3 as al
                import alfred3_interact as ali

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    spec = ali.SequentialSpec("a", "b", nslots=10, name="spec1")
                    exp.plugins.mm = ali.MatchMaker(spec, exp=exp)

                @exp.member
                class Match(ali.WaitingPage):

                    def wait_for(self):
                        group = self.exp.plugins.mm.match()
                        self.exp.plugins.group = group
                        return True

                @exp.member
                class Success(al.Page):

                    def on_first_show(self):
                        group = self.exp.plugins.group
                        role = group.me.role

                        self += al.Text(f"Successfully matched to role: {role}")

        """
        if len(self.groupspecs) > 1:
            msg = (
                "Cannot use the method 'MatchMaker.match()', if there is more than one"
                f" spec. You have {len(self.groupspecs)} specs."
            )
            raise ValueError(msg)

        name = self.groupspecs[0].name
        return self.match_to(name)

    def match_random(self, wait: int = None, nmin: int = None):
        """
        Conducts a match to a random feasible group specification.

        Args:
            wait (int): Number of seconds to wait before actually starting
                a matching effort. Defaults to None, which means
                that there is no waiting time.
            nmin (int): Minimum number of waiting participants that must
                be active before actually starting a matching effort.
                If there are *nmin* participants waiting, matching will
                start even if the waiting time specified in *wait* is
                not over yet. Thus, *nmin* can be used to cut the waiting
                time short. Defaults to None, which means that waiting
                time alone determines when matching will start.
                **Note**: The parameter *nmin* will only take effect, if
                *wait* is not *None*.

        This is the right method, if you wish to randomize participants
        into groups of different sizes.

        .. important::
            :meth:`.match_random` matches randomly among all *feasible*
            specs at the time of matching. A spec is feasible if there
            are enough participants waiting for a group based on that
            spec.

            For instance, let's say you have an :class:`.IndividualSpec`
            and a :class:`.ParallelSpec` with two roles, forming a dyad.
            Four participants
            are scheduled for a session. Participant 1 logs in one minute
            early, Participant 2 is on time, Participant 3 logs in 30
            seconds late, and Participant 4 logs in 2 minutes late. All of
            them will be matched immediately to the :class:`.IndividualSpec`,
            because this spec is always feasible. The :class:`.ParallelSpec`
            on the other hand requires at least two participants to be
            waiting simultaneously.

            This behavior can be adjusted through the arguments *wait*
            and *nmin*. In this example case, you may, for instance,
            specify ``nmin=3`` and ``wait=2*60`` to ensure that there
            is some time for participants to register. Participants 1,
            2, and 3 will be matched to a random spec once Participant 3
            logs in, because *nmin* has been reached. Participant 4 will
            wait for two minutes before being assigned to the individual
            spec. The possible outcomes are: (1) Four individual specs
            (if the randomization does not lead to the formation of a
            dyad), (2) two individual specs and one parallel spec (if a
            dyad is formed in the randomization).

            The method :meth:`.match_chain` may be an even more suitable
            option.

        .. note::
            Randomization into groups of different sizes can be a little
            tricky because large groups may have an unequal (lower)
            chance of being filled at the same rate as smaller groups
            due to participant dropout. We provide the arguments *wait*
            and *nmin* to improve the feasibility of such designs. However,
            the method :meth:`.match_chain` may be an even more suitable
            option.

        Raises:
            NoMatch: If a single matching effort was unsuccesful. This
                exception gets handled by :class:`.WaitingPage` and is
                part of a normal matching process.

        Returns:
            Group: The group object.

        See Also:
            :meth:`.match_chain`

        Examples:
            ::

                import alfred3 as al
                import alfred3_interact as ali

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    spec1 = ali.SequentialSpec("a", "b", nslots=10, name="spec1")
                    spec2 = ali.SequentialSpec("a", "b", nslots=10, name="spec2")
                    exp.plugins.mm = ali.MatchMaker(spec1, spec2, exp=exp)

                @exp.member
                class Match(ali.WaitingPage):

                    def wait_for(self):
                        group = self.exp.plugins.mm.match_random()
                        self.exp.plugins.group = group
                        self.exp.condition = group.data.spec_name
                        return True

                @exp.member
                class Success(al.Page):

                    def on_first_show(self):
                        group = self.exp.plugins.group
                        role = group.me.role
                        cond = self.exp.condition

                        self += al.Text(f"Successfully matched to role '{role}' in condition '{cond}'")

        """
        self.member = self._init_member()

        if self.member.matched:
            return self._get_group(self.member)

        self.member.io.ping()

        start = self._match_start

        if wait is None:
            waited_enough = True
        else:
            waited_enough = time.time() - start >= wait

        if nmin is None:
            enough_members = False
        else:
            enough_members = len(self.waiting_members) >= nmin

        if enough_members or waited_enough:
            random.shuffle(self.groupspecs)
            return self._match_quota(self.groupspecs)

        raise NoMatch

    def match_chain(self, include_previous: bool = True, **spectimes) -> Group:
        """
        Offers prioritized matchmaking based on multiple specs.

        Imagine that you have two :class:`.ParallelSpec` group specifications.
        Spec 1 requires five participants to be active, while Spec 2 only
        requires two.
        You may wish to first try matching based on Spec 1 for some time
        and move on to Spec 2 only when it becomes incresingly clear that
        there are not enough participants present to fill the bigger spec.
        This is what this method is for. You specify a schedule of time
        boxes that are reserved for specific specs.

        Args:
            include_previous (bool): If *True*, earlier specs will still be
                available for matching in later time boxes. If there are
                enough participants waiting for two specs, the MatchMaker
                selects a random spec for matching. Defaults to *True*.
            **spectimes: Specifications of spec timeboxes. Specified by
                ``specname=time``, where *time* is the length of the spec's
                time box in seconds and
                *specname* is the spec's name.
                For the last spec, the value for ``time`` is irrelevant,
                as it will be included until the MatchMaking process
                times out. It should be set to *None*. You can select
                any number of specs, but only specs that are available to
                the MatchMaker.

        Raises:
            NoMatch: If a single matching effort was unsuccesful. This
                exception gets handled by :class:`.WaitingPage` and is
                part of a normal matching process.

        Returns:
            Group: The group object.

        Examples:

            In this example, the MatchMaker will try to form a group
            based on ``spec2`` for the first three minutes of the
            matchmaking process. Once the first active participant
            has waited for more than three minutes without a successfull
            match, ``spec1`` will be included::

                import alfred3 as al
                import alfred3_interact as ali

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    spec1 = ali.ParallelSpec("a", "b", nslots=10, name="spec1")
                    spec2 = ali.ParallelSpec("a", "b", "c", nslots=10, name="spec2")
                    exp.plugins.mm = ali.MatchMaker(spec1, spec2, exp=exp)

                @exp.member
                class Match(ali.WaitingPage):

                    def wait_for(self):
                        group = self.exp.plugins.mm.match_chain(spec2=3*60, spec1=None)
                        self.exp.plugins.group = group
                        self.exp.condition = group.data.spec_name
                        return True

                @exp.member
                class Success(al.Page):

                    def on_first_show(self):
                        group = self.exp.plugins.group
                        role = group.me.role
                        cond = self.exp.condition

                        self += al.Text(f"Successfully matched to role '{role}' in condition '{cond}'")


        """
        # member setip
        self.member = self._init_member()
        if self.member.matched:
            return self._get_group(self.member)
        self.member.io.ping()

        # data setup
        start = self._match_start
        passed_time = time.time() - start

        # remove full specs
        feasible_specs = [spec.name for spec in self.groupspecs if not spec.full(self)]
        spectimes = {
            spec: time for spec, time in spectimes.items() if spec in feasible_specs
        }

        # spec selection setup
        specs = []
        candidate_specs = list(spectimes)
        delays = list(spectimes.values())
        delays = [0, *delays[:-1]]
        previous_delay = 0

        # spec selection
        for i, spec in enumerate(candidate_specs):
            delay = delays[i] + previous_delay
            ready = passed_time >= delay

            try:
                next_delay = delays[i + 1] + delay
                next_spec_ready = passed_time >= next_delay
                over = not include_previous and next_spec_ready
            except IndexError:  # for last spec
                over = False

            if ready and not over:
                specs.append(self.spec_dict[spec])

            previous_delay = delay

        if not specs:
            raise NoMatch

        random.shuffle(specs)
        return self._match_quota(specs)

        # return self._match_quota(specs)

    def match_to(self, name: str) -> Group:
        """
        Matches participants to a group based on a specific spec.

        Args:
            name (str): Name of the spec that should be used.

        Raises:
            NoMatch: If a single matching effort was unsuccesful. This
                exception gets handled by :class:`.WaitingPage` and is
                part of a normal matching process.

        Returns:
            Group: If matching was successful.
        """
        self.member = self._init_member()

        if self.member.matched:
            group = self._get_group(self.member)
            if group.data.spec_name != name:
                msg = (
                    "Member was already matched to a group of spec"
                    f"'{group.data.spec_name}' != '{name}'"
                )
                raise MatchingError(msg)
            return group

        self.member.io.ping()

        return self._match_to(name=name)

    def _get_group(self, member):
        manager = GroupManager(self)
        group = manager.find_one(member.group_id)
        spec = group.data.spec_name
        return self._match_to(spec)

    def _match_to(self, name: str):
        spec = self.spec_dict[name]

        if not self.check_activation() or self.exp.admin_mode:
            return

        self.group = spec._match(self)
        if spec.count:
            try:
                spec.quota.count(self.group, raise_exception=True)
            except AllSlotsFull:
                self.group.deactivate()
                self._full()

        if not self.group:
            return

        self._update_additional_data()
        return self.group

    def toggle_activation(self) -> str:
        """
        Toggles MatchMaker activation.

        An inactive MatchMaker will not conduct any matching.
        Sessions that call a matching method will be aborted.

        Returns:
            str: Returns the new status (active/inactive).
        """
        data = self.io.load()
        data.active = not data.active
        self.io.save(data=data)
        self._data = data

        return "active" if data.active else "inactive"

    def check_quota(self) -> bool:
        """
        Verifies that the MatchMaker has available slots for new
        groups. Aborts the experiment if there are no available slots
        left.

        Raises:
            AllSlotsFull: If the MatchMaker was initialized with
                ``raise_exception_if_full = True`` (default is *False*)
                and the MatchMaker has no available slots.
        """
        if self.exp.admin_mode:
            return True

        if self.quota.full:
            self._full()
            return False
        else:
            return True

    def check_activation(self) -> bool:
        """
        Verifies that the MatchMaker is active and aborts the experiment
        if necessary.

        If the MatchMaker not active, the current experiment session is
        aborted, and the *abort_page* specified on initialization of
        the MatchMaker will be shown to participants.

        The matching methods will check for activation themselves,
        when called. But this may happen at a point in the experiment where
        participants have made considerable progress. So it may be
        useful to manually check for activation at an earlier stage.

        As a general rule of thumb, it may be sensible to check activation
        on first hiding of the landing page of an experiment. This way,
        participants still get a nice welcome screen, but do not invest
        too much time before the experiment is cancelled.

        """
        if self.exp.admin_mode:
            return True

        if self.active:
            return True

        self.log.info("MatchMaking session aborted (MatchMaker inactive).")
        self._abort_exp("matchmaker_inactive", self.inactive_page)

        return False

    def _match_quota(self, specs):
        no_match = False

        for spec in specs:
            if not spec.full(self):
                try:
                    return self._match_to(spec.name)
                except NoMatch:
                    no_match = True

        if no_match:  # only reached if *no* spec lead to successful match
            raise NoMatch

        self._full()

    def _full(self):
        if self.raise_exception_if_full:
            raise AllSlotsFull
        else:
            self._abort_exp("matchmaker_full", self.full_page)

    def _abort_exp(self, reason: str, page):
        full_page_title = "Experiment closed"
        full_page_text = "Sorry, the experiment currently does not accept participants."
        self.exp.abort(
            reason=reason,
            title=full_page_title,
            msg=full_page_text,
            icon="user-check",
            page=page,
        )

    def _validate_specs(self, specs):
        names = {spec.name for spec in specs}
        if len(names) < len(specs):
            raise ValueError("Group specs must have unique names.")

        for spec in specs:
            spec._init_quota(self.exp)

        return specs

    def _init_member(self) -> GroupMember:
        if self.member:
            self.member.io.load()
            return self.member

        member = GroupMember(self)
        member.io.save()
        return member

    def _update_additional_data(self):
        prefix = "interact"
        while prefix in self.exp.adata:
            prefix = "_" + prefix
        self.exp.adata[prefix] = {}
        self.exp.adata[prefix]["groupid"] = self.group.data.group_id
        self.exp.adata[prefix]["role"] = self.member.data.role



from ._util import NoMatch
from .element import Chat, ToggleMatchMakerActivation, ViewMembers
from .match import MatchMaker
from .page import (
    MatchingPage,
    MatchMakerActivation,
    MatchMakerMonitoring,
    MatchTestPage,
    WaitingPage,
)
from .spec import IndividualSpec, ParallelSpec, SequentialSpec

# from .randomizer import GroupRandomizer

# from .quota import GroupQuota



"""
Functionality related to group members.
"""

import datetime
import json
import time
from dataclasses import asdict, dataclass, field
from typing import Iterator, List

from alfred3.data_manager import DataManager as dm
from alfred3.quota import SessionGroup

from ._util import saving_method


@dataclass
class GroupMemberData:
    exp_id: str
    session_id: str
    group_id: str = None
    role: str = None
    created: float = field(default_factory=time.time)
    ping: float = field(default_factory=time.time)
    type: str = "match_member"


class MemberHelper:
    def __init__(self, member):
        self.member = member
        self.sid = self.member.data.session_id
        self.mm = self.member.mm
        self.exp = self.member.exp
        self.saving_method = saving_method(self.exp)


class GroupMemberIO(MemberHelper):
    def __init__(self, member):
        super().__init__(member)

        self.path = self.mm.io.path
        self.db = self.member.exp.db_misc

    @property
    def query(self) -> dict:
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["matchmaker_id"] = self.mm.matchmaker_id
        q["exp_id"] = self.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        return q

    def load(self):
        if self.saving_method == "mongo":
            data = self._load_mongo()
        elif self.saving_method == "local":
            data = self._load_local()
        data = data["members"][self.sid]
        self.member.data = GroupMemberData(**data)

    def _load_local(self) -> dict:
        with open(self.path, encoding="utf-8") as f:
            return json.load(f)

    def _load_mongo(self) -> dict:
        data = self.db.find_one(self.query, projection={f"members.{self.sid}"})
        return data

    def save(self):
        if self.saving_method == "mongo":
            self._save_mongo()
        elif self.saving_method == "local":
            self._save_local()

    def _save_local(self):
        with open(self.path, encoding="utf-8") as f:
            mm = json.load(f)

        sid = self.member.data.session_id
        mm["members"][sid] = asdict(self.member.data)

        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(mm, f, sort_keys=True, indent=4)

    def _save_mongo(self):
        data = {"members": {self.sid: asdict(self.member.data)}}
        self.db.find_one_and_update(self.query, [{"$set": data}])

    def ping(self):
        if saving_method(self.exp) == "local":
            return
        now = time.time()
        data = {"members": {self.sid: {"ping": now}}}
        self.db.find_one_and_update(self.query, [{"$set": data}])
        self.member.data.ping = now


# TODO Manuell deaktivieren für MatchMaker-Chaining
@dataclass
class GroupMemberStatus:
    def __init__(self, member):
        self.member = member
        self.exp = member.exp
        self.status = SessionGroup(sessions=[member.data.session_id])

    @property
    def finished(self) -> bool:
        return self.status.finished(self.exp)

    @property
    def aborted(self) -> bool:
        return self.status.aborted(self.exp)

    @property
    def expired(self) -> bool:
        return self.status.expired(self.exp)

    @property
    def active(self) -> bool:
        return self.status.pending(self.exp)

    @property
    def matched(self) -> bool:
        return self.member.data.group_id is not None

    def ping_expired(self, ping_timeout: int) -> bool:
        now = time.time()
        expired = now - self.member.data.ping > ping_timeout
        return expired

    def print_status(self) -> str:
        if self.active:
            return "active"
        elif self.finished:
            return "finished"
        elif self.aborted:
            return "aborted"
        elif self.expired:
            return "expired"


class GroupMemberExpData(MemberHelper):
    @property
    def db(self):
        return self.exp.db_main

    @property
    def query(self) -> dict:
        d = {}
        d["exp_id"] = self.exp.exp_id
        d["exp_session_id"] = self.sid
        d["type"] = dm.EXP_DATA

        return d

    def _load_local(self) -> dict:
        dt = dm.EXP_DATA
        directory = self.exp.config.get("local_saving_agent", "path")
        directory = self.exp.subpath(directory)
        cursor = dm.iterate_local_data(dt, directory)
        session = (data for data in cursor if data["exp_session_id"] == self.sid)
        return next(session, None)

    def load(self, projection=None) -> dict:
        if self.saving_method == "mongo":
            data = self.db.find_one(self.query, projection=projection)
            data.pop("_id")

        elif self.saving_method == "local":
            data = self._load_local()

            if isinstance(projection, list):
                data = {k: v for k, v in data.items() if k in projection}
            elif isinstance(projection, dict):
                data = {k: v for k, v in data.items() if projection.get(k, True)}

        return data

    @property
    def start_time_unix(self) -> float:
        data = self.load(["exp_start_time"])
        return data["exp_start_time"]


class GroupMemberInfo(MemberHelper):
    def __init__(self, member):
        super().__init__(member)
        self.expdata = self.member.expdata
        self._start_time_unix = self.expdata.start_time_unix

    @property
    def start_time_unix(self) -> float:
        if self._start_time_unix:
            start_time = self._start_time_unix
        else:
            start_time = self.expdata.start_time_unix
            self._start_time_unix = start_time

        return start_time

    @property
    def start_time(self) -> str:
        """
        str: Session start-time in human-readbale format (H:M:S)
        """
        if self.start_time_unix:
            return time.strftime("%H:%M:%S", time.localtime(self.start_time_unix))
        else:
            return "None"

    @property
    def start_day(self) -> str:
        """
        str: Session start-day in human-readbale format (yyyy-mm-dd)
        """
        date = datetime.date.fromtimestamp(self.start_time_unix)
        return date.isoformat()

    @property
    def last_move(self) -> str:
        moves = self.member.move_history
        if moves:
            t = moves[-1]["hide_time"]
            time_print = time.strftime("%H:%M:%S", time.localtime(t))
            return time_print

        return None

    @property
    def last_save(self) -> float:
        projection = {"exp_save_time": True, "_id": False}
        data = self.member.expdata.load(projection)
        return data["exp_save_time"]

    @property
    def last_page(self) -> str:
        try:
            return self.member.move_history[-1]["target_page"]
        except IndexError:
            return "-landing page-"


class GroupMember:
    """
    The group member object grants access to a member's experiment data.

    The group member object's most important job is to provide easy
    access to the member's experiment data through the following
    attributes. They provide the same objects as the corresponding
    attributes of the :class:`alfred3.experiment.ExperimentSession`
    object:

    - :attr:`.values`
    - :attr:`.session_data`
    - :attr:`.metadata`
    - :attr:`.client_data`
    - :attr:`.adata`
    """

    def __init__(self, matchmaker, **data):

        self.mm = matchmaker
        self.exp = self.mm.exp

        self.data = GroupMemberData(**self._prepare_data(data))
        self.io = GroupMemberIO(self)
        self.status = GroupMemberStatus(self)
        self.expdata = GroupMemberExpData(self)
        self.info = GroupMemberInfo(self)

    def _prepare_data(self, data: dict) -> dict:
        exp_id = data.get("exp_id", None)
        sid = data.get("session_id", None)

        data["exp_id"] = exp_id if exp_id is not None else self.exp.exp_id
        data["session_id"] = sid if sid is not None else self.exp.session_id

        data.pop("_id", None)
        return data

    @property
    def role(self) -> str:
        """
        str: The member's role.
        """
        return self.data.role

    @property
    def matched(self) -> bool:
        """
        bool: Indicates whether the member is associated with a group.
        """
        return self.status.matched

    @property
    def group_id(self) -> str:
        """
        str: If the member is associated with a group, this property
        returns the group id.
        """
        return self.data.group_id

    @property
    def values(self) -> dict:
        """
        dict: Flat dictionary of input element values.

        Gives access to the member's experiment inputs.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.values`. The keys are
            the names of input elements in the member's experiment session.
            The values are the user inputs.
        """
        projection = {}
        projection.update({key: False for key in dm._client_data_keys})
        projection.update({key: False for key in dm._metadata_keys})

        data = self.expdata.load(projection)
        return dm.flatten(data)

    @property
    def session_data(self) -> dict:
        """
        dict: Full dictionary of experiment session data.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.session_data`.
        """
        return self.expdata.load()

    @property
    def client_data(self) -> dict:
        """
        dict: Dictionary of client data.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.client_data`.
        """
        client_data = list(dm._client_data_keys)
        return self.expdata.load(client_data)

    @property
    def metadata(self) -> dict:
        """
        dict: Dictionary of experiment metadata.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.metadata`.
        """
        metadata = list(dm._metadata_keys)
        return self.expdata.load(metadata)

    @property
    def move_history(self) -> dict:
        """
        dict: Dictionary of movement history.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.move_history`.
        """
        data = self.expdata.load(["exp_move_history"])
        return data["exp_move_history"]

    @property
    def adata(self) -> dict:
        """
        dict: Dictionary of additional data.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.adata`.
        """
        data = self.expdata.load(["additional_data"])
        return data["additional_data"]

    @property
    def additional_data(self) -> dict:
        """
        dict: Alias for :attr:`.adata`
        """
        return self.adata

    def __repr__(self) -> str:
        gid = self.data.group_id[-4:] if self.data.group_id is not None else "-"
        repr = (
            f"{type(self).__name__}(role='{self.data.role}', "
            f"session_id='{self.data.session_id[-4:]}', "
            f"group='{gid}', start='{self.info.start_time}')"
        )

        return repr

    def __eq__(self, other) -> bool:
        class_equal = type(self) == type(other)
        session_id_equal = self.data.session_id == other.data.session_id

        return class_equal and session_id_equal


class MemberManager:
    def __init__(self, matchmaker):
        self.mm = matchmaker
        self.exp = self.mm.exp
        self.db = self.exp.db_misc
        self._active_sessions_projection = None
        self._last_update = None
        self.method = saving_method(self.exp)

    @property
    def query_mm(self) -> dict:
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["matchmaker_id"] = self.mm.matchmaker_id
        q["exp_id"] = self.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        return q

    @property
    def query_exp(self) -> dict:
        q = {}
        q["type"] = dm.EXP_DATA
        q["exp_id"] = self.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        return q

    def query_active_sessions(self, sessions: List[str] = None) -> dict:
        q = self.query_exp
        q["exp_finished"] = False
        q["exp_aborted"] = False

        if sessions is not None:
            q["exp_session_id"] = {"$in": sessions}

        return q

    def query_finished_sessions(self, sessions: List[str] = None) -> dict:
        q = self.query_exp
        q["exp_finished"] = True
        q["exp_aborted"] = False

        if sessions is not None:
            q["exp_session_id"] = {"$in": sessions}

        return q

    def find_finished_sessions(self, sessions: List[str] = None) -> Iterator[str]:
        if self.method == "local":
            return self._find_finished_sessions_local(sessions)
        elif self.method == "mongo":
            return self._find_finished_sessions_mongo(sessions)

    def find_active_sessions(self, sessions: List[str] = None) -> Iterator[str]:
        if self.method == "local":
            return self._find_active_sessions_local(sessions)
        elif self.method == "mongo":
            return self._find_active_sessions_mongo(sessions)

    def _find_active_sessions_local(self, sessions: List[str] = None) -> Iterator[str]:
        for member in self.members():
            if member.status.active:
                yield member.data.session_id

    def _find_finished_sessions_mongo(
        self, sessions: List[str] = None
    ) -> Iterator[str]:
        q = self.query_finished_sessions(sessions)
        cursor = self.exp.db_main.find(q, projection=["exp_session_id"])
        for sessiondata in cursor:
            status = SessionGroup([sessiondata["exp_session_id"]])
            if status.finished(self.exp):
                yield sessiondata["exp_session_id"]

    def _find_active_sessions_mongo(self, sessions: List[str] = None) -> Iterator[str]:
        q = self.query_active_sessions(sessions)
        cursor = self.exp.db_main.find(q, projection=["exp_session_id"])
        for sessiondata in cursor:
            status = SessionGroup([sessiondata["exp_session_id"]])
            if status.pending(self.exp):
                yield sessiondata["exp_session_id"]

    def _find_finished_sessions_local(self, sessions: List[str]) -> Iterator[str]:
        for m in self.members():
            if m.status.finished and m.data.session_id in sessions:
                yield m.data.session_id

    def active_sessions_projection(self, cache_length: int = 1) -> dict:
        now = time.time()

        if self._last_update and self._last_update - now < cache_length:
            return self._active_sessions_projection
        else:
            p = {f"members.{sid}": True for sid in self.find_active_sessions()}
            p.update({"_id": False})
            self._active_sessions_projection = p
            self._last_update = now

    def active(self) -> GroupMember:
        if self.method == "local":
            return self._active_local()

        elif self.method == "mongo":
            return self._active_mongo()

    def _active_local(self) -> Iterator[GroupMember]:
        for m in self.members():
            if m.status.active:
                yield m

    def _active_mongo(self) -> Iterator[GroupMember]:
        q = self.query_mm
        p = self.active_sessions_projection()

        data = self.db.find_one(filter=q, projection=p)
        data = data["members"].values()

        for mdata in data:
            yield GroupMember(matchmaker=self.mm, **mdata)

    def waiting(self, ping_timeout: int) -> Iterator[GroupMember]:
        for m in self.unmatched():
            if not m.status.ping_expired(ping_timeout):
                yield m

    def members(self) -> Iterator[GroupMember]:
        data = self.mm.io.load().members.values()
        for mdata in data:
            yield GroupMember(matchmaker=self.mm, **mdata)

    def unmatched(self) -> Iterator[GroupMember]:
        for m in self.active():
            if not m.status.matched:
                yield m

    def matched(self) -> Iterator[GroupMember]:
        for m in self.active():
            if m.status.matched:
                yield m

    def find(self, sessions: List[str]) -> Iterator[GroupMember]:
        if self.method == "local":
            return self._find_local(sessions)

        elif self.method == "mongo":
            return self._find_mongo(sessions)

    def _find_local(self, sessions: List[str]) -> Iterator[GroupMember]:
        data = self.mm.io.load().members.values()
        for mdata in data:
            if mdata["session_id"] in sessions:
                yield GroupMember(matchmaker=self.mm, **mdata)

    def _find_mongo(self, sessions: List[str]) -> Iterator[GroupMember]:
        q = self.query_mm
        p = {f"members.{sid}": True for sid in sessions}
        p.update({"_id": False})

        data = self.db.find_one(filter=q, projection=p)
        data = data["members"].values()

        for mdata in data:
            yield GroupMember(matchmaker=self.mm, **mdata)
